from dataclasses import dataclass
from typing import Optional, List

import rich
from documented import DocumentedError
from more_itertools import first
from typer import Argument

from jirajumper.cache.cache import JeevesJiraContext


def comment(
    context: JeevesJiraContext,
    text: str = Argument(...),
):
    """Comment an issue."""
    issue = context.obj.current_issue
    jira = context.obj.jira

    jira.add_comment(
        issue=issue.key,
        body=text,
    )

    rich.print('💬 Commented.')



from enum import Enum
from types import MappingProxyType
from typing import List, Set

import rich
from jira import JIRA, Issue
from typer import Argument

from jirajumper.cache.cache import JeevesJiraContext
from jirajumper.commands.select import normalize_issue_specifier


class LinkType(str, Enum):   # noqa: WPS600
    """
    Supported Jira link types.

    # FIXME Can we deduce that dynamically and store in cache?
    """

    SPLIT_FROM = 'split-from'
    SPLIT_TO = 'split-to'

    RELATES_TO = 'relates-to'

    IS_CAUSED_BY = 'is-caused-by'
    CAUSES = 'causes'

    IS_DUPLICATED_BY = 'is-duplicated-by'
    DUPLICATES = 'duplicates'

    DEPENDS_ON = 'deps'
    DEPENDED_ON_BY = 'dep'

    IS_BLOCKED_BY = 'is-blocked-by'
    BLOCKS = 'blocks'

    REMOVE = 'remove'
    CONFLUENCE = 'confluence'
    LIST = 'list'

    @property
    def jira_name(self):
        """Jira native name of the transition."""
        return self.name.lower().replace('_', ' ')


def remove_link(
    current_issue: Issue,
    issue_keys: Set[str],
    link_type: LinkType,
    jira: JIRA,
):
    """Remove links to a number of issues."""
    for existing_link in current_issue.fields.issuelinks:
        linked_issue = getattr(
            existing_link, 'outwardIssue', None,
        ) or existing_link.inwardIssue

        if linked_issue.key in issue_keys:
            jira.delete_issue_link(existing_link.id)
            rich.print(
                f'{current_issue.key} and {linked_issue.key} '
                f'are no longer connected.',
            )


def link_confluence(
    current_issue: Issue,
    issue_keys: Set[str],
    link_type: LinkType,
    jira: JIRA,
):
    """Remove links to a number of issues."""
    raise NotImplementedError('Linking to Confluence is not yet implemented.')


def list_links(
    current_issue: Issue,
    issue_keys: Set[str],
    link_type: LinkType,
    jira: JIRA,
):
    """Remove links to a number of issues."""
    rich.print('Issue links:')
    for existing_link in current_issue.fields.issuelinks:
        linked_issue = getattr(
            existing_link, 'outwardIssue', None,
        ) or existing_link.inwardIssue

        if linked_issue.key in issue_keys:
            rich.print(
                f'  * {current_issue.key} {existing_link} {linked_issue.key}',
            )


def link_default(
    current_issue: Issue,
    issue_keys: Set[str],
    link_type: LinkType,
    jira: JIRA,
):
    """Create a link between issues."""
    for issue_key in issue_keys:
        jira.create_issue_link(
            type=link_type.jira_name,
            inwardIssue=current_issue.key,
            outwardIssue=issue_key,
        )
        rich.print(
            f'* {current_issue} {link_type} {issue_key}.',
        )


LINK_MANAGERS = MappingProxyType({
    LinkType.CONFLUENCE: link_confluence,
    LinkType.REMOVE: remove_link,
    LinkType.LIST: list_links,
})


def link(
    context: JeevesJiraContext,
    link_type: LinkType,
    specifiers: List[str] = Argument(None),  # noqa: WPS404, B008
):
    """Link current issue to some other issue."""
    parent_issue = context.obj.current_issue
    jira = context.obj.jira

    issue_keys = {
        normalize_issue_specifier(
            client=jira,
            specifier=specifier,
            current_issue_key=parent_issue.key,
        )
        for specifier in specifiers or []
    }

    linker = LINK_MANAGERS.get(
        link_type,
        link_default,
    )

    return linker(
        current_issue=parent_issue,
        issue_keys=issue_keys,
        link_type=link_type,
        jira=jira,
    )



from dataclasses import dataclass
from typing import Dict, Optional

import rich
from documented import DocumentedError
from jira import JIRA, JIRAError

from jirajumper import default_options
from jirajumper.cache.cache import JeevesJiraContext
from jirajumper.fields import JiraFieldsRepository


@dataclass
class JIRAUpdateFailed(DocumentedError):
    """
    Cannot update a JIRA issue 🙁.

    Errors:
    {self.formatted_errors}
    """

    errors: Dict[str, str]
    fields: JiraFieldsRepository

    @property
    def formatted_errors(self) -> str:
        """Format the error list received from JIRA."""
        field_per_resolved_jira_name = {
            field.jira_name: field
            for field in self.fields
        }

        error_by_field = [
            (field_per_resolved_jira_name[jira_name], error_message)
            for jira_name, error_message in self.errors.items()
        ]

        return '\n'.join([
            f'  - {field.human_name}\n      {error_message}'
            for field, error_message in error_by_field
        ])


def assign(jira: JIRA, key: str, assignee: str):
    """Assign issue to a person."""
    rich.print(f'Assigning {key} to {assignee}...')
    jira.assign_issue(
        issue=key,
        assignee=assignee,
    )

    rich.print('Assigned!')


def update(
    context: JeevesJiraContext,
    assignee: Optional[str] = default_options.ASSIGNEE,
    **options: str,
):
    """
    Update the selected JIRA issue.

    Use `jj jump` to select the issue to update.
    """
    fields_and_values = context.obj.fields.match_options(options)

    rich.print('Updating:')
    for print_field, human_value in fields_and_values:
        rich.print(f'  - {print_field.human_name} ≔ {human_value}')

    issue_fields = dict([
        store_field.store(human_value=human_value)
        for store_field, human_value in fields_and_values
    ])

    try:
        context.obj.current_issue.update(issue_fields)
    except JIRAError as err:
        raise JIRAUpdateFailed(
            errors=err.response.json().get('errors', {}),
            fields=context.obj.fields,
        ) from err

    if assignee is not None:
        assign(
            jira=context.obj.jira,
            key=context.obj.current_issue.key,
            assignee=assignee,
        )

    rich.print('Updated!')



from dataclasses import dataclass
from typing import Optional, List

import rich
from documented import DocumentedError
from more_itertools import first
from typer import Argument

from jirajumper.cache.cache import JeevesJiraContext


@dataclass
class NoTransitionFound(DocumentedError):
    """
    No transition found!

        - Source: {self.source_status}
        - Destination: {self.destination_status}

    Transitive transitions are not yet supported ☹
    """

    source_status: str
    destination_status: str


def status(
    context: JeevesJiraContext,
    status_values: Optional[List[str]] = Argument(None),
):
    """Get or set issue status."""
    issue = context.obj.current_issue
    jira = context.obj.jira

    if status_values is None:
        rich.print(issue.fields.status)
        return

    for status in status_values:
        transitions = jira.transitions(issue=issue.key)

        suitable_transitions = [
            transition['name']
            for transition in transitions
            if transition['to']['name'].lower() == status.lower()
        ]

        try:
            target_transition = first(suitable_transitions)
        except ValueError as err:
            raise NoTransitionFound(
                source_status=issue.fields.status.name,
                destination_status=status,
            ) from err

        jira.transition_issue(
            issue=issue.key,
            transition=target_transition,
        )
        rich.print(f'✔️ Status is now [bold]{status}[/bold].')



from typing import Optional

import rich
from jira import JIRAError
from typer import Option

from jirajumper import default_options
from jirajumper.cache.cache import JeevesJiraContext
from jirajumper.commands.clone import clone
from jirajumper.commands.link import LinkType, link
from jirajumper.commands.select import jump
from jirajumper.commands.update import JIRAUpdateFailed


def fork(
    context: JeevesJiraContext,
    link_type: LinkType = Option(LinkType.BLOCKS, '--link'),
    stay: bool = False,
    assignee: Optional[str] = default_options.ASSIGNEE,
    summary: str = default_options.SUMMARY,
    **options: str,
):
    """Fork a JIRA issue."""
    child_issue = clone(
        context=context,
        stay=True,
        assignee=assignee,
        summary=summary,
        **options,
    )

    link(
        context=context,
        link_type=link_type,
        specifiers=[child_issue.key],
    )

    if not stay:
        jump(
            context=context,
            specifier=child_issue.key,
        )



from typing import Dict

import rich

from jirajumper.cache.cache import JeevesJiraContext
from jirajumper.fields import JiraFieldsRepository
from jirajumper.fields.field import ResolvedField


def generate_jql(
    fields: JiraFieldsRepository,
    options: Dict[str, str],
):
    """Generate JQL string."""
    fields_and_values = fields.match_options(options)

    field: ResolvedField
    expressions = [
        field.to_jql(expression)
        for field, expression in fields_and_values
    ]

    return ' AND '.join(expressions)


def list_issues(
    context: JeevesJiraContext,
    **options,
):
    """List JIRA issues by criteria."""
    jql = generate_jql(
        fields=context.obj.fields,
        options=options,
    )
    context.obj.logger.info('JQL: `%s`', jql)
    issues = context.obj.jira.search_issues(jql, maxResults=None)

    for issue in issues:
        rich.print(
            '* {key} [i]({status} / {assignee})[/i] {summary}'.format(
                key=issue.key,
                status=issue.fields.status,
                assignee=issue.fields.assignee,
                summary=issue.fields.summary,
            ),
        )



import json
import os
from pathlib import Path

from jirajumper.cache.cache import JiraCache


def project_path() -> Path:
    """Find project directory."""
    current_dir = Path.cwd()

    if (current_dir / 'pyproject.toml').exists():
        return current_dir

    raise ValueError('This is not a project dir!')


def construct_cache_file_path() -> Path:
    """Path to jirajumper cache file."""
    return Path(os.getenv(
        'JIRAJUMPER_CACHE',
        Path.home() / '.cache/jirajumper',
    )) / 'jirajumper.json'



import html
import textwrap
from types import MappingProxyType

import graphviz

from jirajumper.cache.cache import JeevesJiraContext
from jirajumper.client import issue_url
from jirajumper.commands.list_issues import generate_jql

DEFAULT_COLOR_MAP = MappingProxyType({
    'Code Review': '#BD34D1',  # Pink
    'In Review': '#BD34D1',  # Pink
    'Refinement': '#897A5F',  # Brown
    'In Progress': '#F7C325',  # Yellow
    # 'Queued': '#AC6363',      # Crimson
    'Rejected': '#AC6363',  # Crimson
    'Queued': '#D3455B',  # Red
    # 'Draft': '#CCCCCC',       # Gray
    # 'Draft': '#E8833A',  # Orange
    'Draft': '#EEEEEE',  # Light Gray
    # 'Closed': '#E8833A',      # Orange
    'Closed': '#CCCCCC',  # Gray
    'Put On Ice': '#2C88D9',  # Blue
    'QA': '#730FC3',  # Purple
    'In Testing': '#730FC3',  # Purple
    # 'Done': '#207868',        # Green
    'Done': '#1AAE9F',  # Mint
})


def graph(
    context: JeevesJiraContext,
    **options,
):
    """List JIRA issues by criteria."""
    jql = generate_jql(
        fields=context.obj.fields,
        options=options,
    )

    logger = context.obj.logger
    logger.info('JQL: `%s`', jql)

    jira = context.obj.jira
    issues = jira.search_issues(jql, maxResults=None)

    graph = graphviz.Digraph(
        comment='Jira task links',
        graph_attr={
            'rankdir': 'LR',
        }
    )
    for issue in issues:
        color = DEFAULT_COLOR_MAP.get(
            issue.fields.status.name,
            'white',
        )

        wrapped_summary = '<br/>'.join(
            textwrap.wrap(
                html.escape(issue.fields.summary.replace('"', '')),
                width=20,
            )
        )

        if (
            issue.fields.issuetype.name == 'Epic' and
            issue.fields.status.statusCategory.name != 'Done'
        ):
            logger.info('Retrieving children for epic %s...', issue)
            epic_children = jira.search_issues(
                f'"Epic Link" = {issue.key}',
                maxResults=None,
            )
            total_count = len(epic_children)
            done_count = len([
                issue
                for issue in epic_children
                if issue.fields.status.statusCategory.name == 'Done'
            ])
            progress = f'<I>[{done_count} of {total_count}]</I>'
        else:
            progress = ''

        label = textwrap.dedent(f'''
            <TABLE BGCOLOR="{color}" BORDER="0" CELLBORDER="1" CELLSPACING="0">
                <TR><TD ALIGN="left"><B>{issue.key}</B> <I>{issue.fields.issuetype}</I></TD></TR>
                <TR><TD ALIGN="left">{wrapped_summary}</TD></TR>
                <TR><TD ALIGN="left">{issue.fields.assignee}</TD></TR>
                <TR><TD ALIGN="left">{issue.fields.status} {progress}</TD></TR>
            </TABLE>
        ''')

        graph.node(
            issue.key,
            f'<{label}>',
            shape='none',
            href=issue_url(context.obj.jira.server_url, issue.key),
            target='_blank',
        )

        links = issue.fields.issuelinks
        for link in links:
            linked_issue = getattr(link, 'inwardIssue', None)
            if not linked_issue:
                continue

            if (
                linked_issue.fields.status.name == 'Put On Ice' and
                issue.fields.status.name in {'In Progress', 'Queued'}
            ):
                edge_options = {'color': '#D3455B', 'fontcolor': '#D3455B'}
            else:
                edge_options = {}

            graph.edge(
                linked_issue.key,
                issue.key,
                link.type.name,
                **edge_options,
            )

    graph.render(
        '/tmp/jeeves-jira-roadmap',
        format='png',
        view=True,
    )



from typing import Optional

from jira import JIRAError

from jirajumper import default_options
from jirajumper.cache.cache import JeevesJiraContext
from jirajumper.commands.select import jump
from jirajumper.commands.update import JIRAUpdateFailed, assign


def clone(
    context: JeevesJiraContext,
    stay: bool = False,
    assignee: Optional[str] = default_options.ASSIGNEE,
    summary: str = default_options.SUMMARY,
    **options: str,
):
    """Clone a JIRA issue."""
    options.update({
        'summary': summary,
    })

    parent_issue = context.obj.current_issue
    parent_issue_fields = dict(
        field.store(field.retrieve(issue=parent_issue))
        for field in context.obj.fields
        if field.is_writable()
    )

    resolved_fields = context.obj.fields.match_options(options)

    update_fields = dict(
        field.store(human_value)
        for field, human_value in resolved_fields
    )

    new_issue_fields = {
        **parent_issue_fields,
        **update_fields,
    }

    try:
        issue = context.obj.jira.create_issue(fields=new_issue_fields)
    except JIRAError as err:
        raise JIRAUpdateFailed(
            errors=err.response.json().get('errors', {}),
            fields=context.obj.fields,
        ) from err

    if not assignee:
        assignee = getattr(parent_issue.fields.assignee, 'displayName', None)

    if assignee:
        assign(
            jira=context.obj.jira,
            key=issue.key,
            assignee=assignee,
        )

    if not stay:
        jump(
            context=context,
            specifier=issue.key,
        )

    return issue



from typer import Option, Argument

ASSIGNEE = Option(
    None,
    help='Assignee display name or email address. Supports fuzzy search.',
)

SUMMARY = Argument(
    ...,
    help='Issue summary.'
)



import logging
import sys
from enum import Enum
from functools import partial
from itertools import filterfalse
from pathlib import Path

import click
import rich
from jirajumper.commands.comment import comment
from rich.traceback import install
from typer import Context, Option, Typer
from typer.core import TyperArgument, TyperCommand

from jirajumper.cache.cache import GlobalOptions, field_key_by_name
from jirajumper.client import jira
from jirajumper.commands.clone import clone
from jirajumper.commands.fork import fork
from jirajumper.commands.graph import graph
from jirajumper.commands.link import link
from jirajumper.commands.list_issues import list_issues
from jirajumper.commands.select import jump
from jirajumper.commands.status import status
from jirajumper.commands.update import update
from jirajumper.fields import FIELDS, JiraField, JiraFieldsRepository
from jirajumper.models import OutputFormat

app = Typer(
    help='Manage JIRA issues.',
    no_args_is_help=True,
)


class LogLevel(str, Enum):  # noqa: WPS600
    """Available logging levels."""

    DEBUG = 'debug'
    ERROR = 'error'


def exception_handler(exception_type, exception, traceback):
    """Custom exception handler to look more civilized than we are."""
    exception_name = exception_type.__name__
    exception_text = str(exception)
    rich.print(
        f'[red][bold]{exception_name}:[/bold][/red] {exception_text}',
    )


@app.callback()
def global_options(
    context: Context,
    format: OutputFormat = Option(
        OutputFormat.PRETTY,
        help='Format to print the data in',
    ),
    cache_path: Path = Option(
        default=Path.home() / '.cache/jirajumper/jirajumper.json',
        envvar='JIRAJUMPER_CACHE_PATH',
        help='Path to the JSON file where jirajumper will store its cache.',
    ),
    log_level: LogLevel = Option(   # noqa: WPS404, B008
        LogLevel.ERROR,
        help=(
            'Log message level: `debug` (to print all debug messages and '
            'exception tracebacks), or `error` (to only log critical errors).'
        ),
    ),
):
    """Configure global options valid for most of jeeves-jira commands."""
    logger = logging.getLogger('jj')

    if log_level == LogLevel.DEBUG:
        install(show_locals=False)
        logger.setLevel(logging.DEBUG)
    else:
        sys.excepthook = exception_handler
        logger.setLevel(logging.ERROR)

    client = jira()
    key_by_name = field_key_by_name(
        jira=client,
    )

    resolved_fields = map(
        partial(
            JiraField.resolve,
            field_key_by_name=key_by_name,
        ),
        FIELDS,
    )

    context.obj = GlobalOptions(
        logger=logger,
        output_format=format,
        jira=jira(),
        fields=JiraFieldsRepository(resolved_fields),
        cache_path=cache_path,
    )


class AutoOptionsCommand(TyperCommand):
    writable_only = False
    mutable_only = False

    def __init__(self, **kwargs):
        fields = FIELDS

        if self.mutable_only:
            fields = fields.mutable()

        if self.writable_only:
            fields = fields.writable()

        custom_options = [
            click.Option(
                ['--{option_name}'.format(
                    option_name=field.human_name.replace("_", "-"),
                )],
                help=field.description,
            )
            for field in fields
        ]

        existing_params = list(filterfalse(
            lambda existing_param: (
                isinstance(existing_param, TyperArgument) and
                existing_param.name == 'options'
            ),
            kwargs.get('params', []),
        ))

        kwargs.update(
            params=existing_params + custom_options,
        )

        super().__init__(**kwargs)


class CloneCommand(AutoOptionsCommand):
    writable_only = True


class UpdateCommand(AutoOptionsCommand):
    mutable_only = True


app.command()(jump)

app.command(
    cls=CloneCommand,
    context_settings={
        'ignore_unknown_options': True,
    },
)(clone)

app.command(
    cls=CloneCommand,
    context_settings={
        'ignore_unknown_options': True,
    },
)(fork)

app.command(
    cls=UpdateCommand,
    context_settings={
        'ignore_unknown_options': True,
    },
)(update)


app.command(
    cls=AutoOptionsCommand,
    context_settings={
        'ignore_unknown_options': True,
    },
    name='list',
)(list_issues)

app.command(
    cls=AutoOptionsCommand,
    context_settings={
        'ignore_unknown_options': True,
    },
    name='graph',
)(graph)

app.command()(link)
app.command()(status)
app.command()(comment)



import operator
import re
from dataclasses import asdict, dataclass
from typing import Optional, Protocol, Tuple, TypeVar, Union

from jira import Issue

from jirajumper.models import (
    FieldByName,
    FieldKeyByName,
    HumanValue,
    JiraValue,
    NotImplementedType,
)

AnyType = TypeVar('AnyType')


class ToJira(Protocol):
    """Convert a human-readable value to JIRA field."""

    def __call__(self, human_value: HumanValue) -> JiraValue:
        """Convert a human-readable value to JIRA field."""
        raise NotImplementedError()


class FromJira(Protocol):
    """Convert a native JIRA field value to a human readable one."""

    def __call__(self, jira_value: JiraValue) -> HumanValue:
        """Convert a native JIRA field value to a human readable one."""
        raise NotImplementedError()


class ToJQL(Protocol):
    """Construct a JQL expression from a raw argument value."""

    def __call__(self, expression: str) -> str:
        """Construct a JQL expression from a raw argument value."""
        raise NotImplementedError()


def identity(any_value: AnyType) -> AnyType:
    """Identity function."""
    return any_value


@dataclass(frozen=True)
class JiraField:
    """Transformations related to a particular JIRA field."""

    jira_name: Union[str, FieldByName]
    human_name: str
    description: str

    is_mutable: bool = True

    jql_name: Optional[str] = None
    to_jira: Union[ToJira, NotImplementedType] = identity
    from_jira: FromJira = identity

    def retrieve(self, issue: Issue):
        """Retrieve the native field value from given issue."""
        return self.from_jira(
            operator.attrgetter(self.jira_name)(issue.fields),
        )

    def store(self, human_value: HumanValue) -> Tuple[str, JiraValue]:
        """Convert the readable value into JIRA native form."""
        return self.jira_name, self.to_jira(human_value)

    def is_writable(self):
        """Find out if this field is writable."""
        return self.to_jira is not NotImplemented

    def resolve(self, field_key_by_name: FieldKeyByName) -> 'ResolvedField':
        """Resolve jira_name."""
        if isinstance(self.jira_name, str):
            unresolved_name = self.jira_name
            resolved_name = self.jira_name

        elif isinstance(self.jira_name, FieldByName):
            unresolved_name = self.jira_name.readable_name
            resolved_name = field_key_by_name[unresolved_name]

        else:
            raise ValueError(
                f'`{self.jira_name}` is not a valid JIRA field name.',
            )

        field_dict = {
            **asdict(self),
            **{
                'unresolved_jira_name': unresolved_name,
                'jira_name': resolved_name,
            },
        }

        return ResolvedField(**field_dict)


def _jql_operator(is_multiple: bool, is_positive: bool) -> str:
    return {
        False: {
            False: '!=',
            True: '=',
        },
        True: {
            False: 'NOT IN',
            True: 'IN',
        },
    }[is_multiple][is_positive]


@dataclass(frozen=True)
class ResolvedField(JiraField):
    """JIRA field description with resolved field name."""

    unresolved_jira_name: Optional[str] = None

    def to_jql(self, expression: str) -> str:   # noqa: WPS210
        """Convert human readable expression to JQL."""
        minus, pattern = re.match('(-*)(.+)', expression).groups()
        is_positive = not minus

        search_values = list(map(
            str.strip,
            pattern.split(','),
        ))
        is_multiple = len(search_values) > 1

        jql_operator = _jql_operator(
            is_multiple=is_multiple,
            is_positive=is_positive,
        )

        jql_values = ', '.join(
            f'"{search_value}"'
            for search_value in search_values
        )

        if is_multiple:
            jql_values = f'({jql_values})'

        field_name = self.jql_name or self.unresolved_jira_name
        if ' ' in field_name:
            field_name = f'"{field_name}"'

        return f'{field_name} {jql_operator} {jql_values}'



from dataclasses import dataclass
from typing import Optional

from documented import DocumentedError


@dataclass
class MissingJiraCredentials(DocumentedError):
    """
    JIRA credentials are not properly configured.

    Normally, it is recommended to configure your JIRA settings as environment
    variables - say, in `~/.profile`.

        export JIRA_SERVER='{self.server_safe}'
        export JIRA_USERNAME='{self.username_safe}'
        export JIRA_TOKEN='{self.token_safe}'

    See https://yeti.sh/jira/token about how to create an API token for your
    Atlassian account.
    """

    server: Optional[str]
    username: Optional[str]
    token: Optional[str]

    def safe(self, credential: Optional[str]) -> str:
        if credential:
            return '(something)'
        else:
            return '(not provided)'

    @property
    def server_safe(self):
        """Safe server."""
        return self.safe(self.server)

    @property
    def username_safe(self):
        """Safe username."""
        return self.safe(self.username)

    @property
    def token_safe(self):
        """Safe token."""
        return self.safe(self.token)



from typing import Dict, List, Optional, Tuple

from more_itertools import first

from jirajumper.fields.field import JiraField


class JiraFieldsRepository(List[JiraField]):
    """List of supported JIRA issue fields."""

    def find_by_jira_name(self, jira_name: str) -> Optional[JiraField]:
        """Find a field by JIRA name."""
        return first(
            filter(
                lambda field: field.jira_name == jira_name,
                self,
            ), None,
        )

    def writable(self) -> 'JiraFieldsRepository':
        """Show only fields that are writable."""
        return JiraFieldsRepository(filter(
            JiraField.is_writable,
            self,
        ))

    def mutable(self) -> 'JiraFieldsRepository':
        """Show only fields that are mutable."""
        return JiraFieldsRepository(filter(
            lambda field: field.is_writable and field.is_mutable,
            self,
        ))

    def match_options(
        self,
        options: Dict[str, str],
    ) -> List[Tuple[JiraField, str]]:
        """
        Match fields in the repo with CLI options provided by the user.

        Returns a list of `(JiraField, str)` pairs, where `str` is the value
        assigned to this field by the user.
        """
        return [
            (field, options[field.human_name])
            for field in self
            if options.get(field.human_name)
        ]
